import numpy as np
import pandas as pd
from svvamp.utils.misc import preferences_ut_to_preferences_borda_ut


def cvr_to_preferences_ut(file_name):
    df = pd.read_csv(filepath_or_buffer=file_name, sep=',', index_col=False)
    last_rank = int(df.columns[-1][4:])  # E.g. if the last column is named 'rank3' => last_rank = 3.
    # Drop useless columns
    df.drop(columns=df.columns[:-last_rank], inplace=True)
    # Replace special values with NaN
    df.replace(to_replace=r'^(WRITE-IN)|(writein)|(Write-In)|(Write-in)|(skipped)|(overvote).*',
               value=np.nan, regex=True, inplace=True)
    # Compute the set of candidates
    set_candidates = set()
    for rank in range(1, last_rank + 1):
        column_name = 'rank{}'.format(rank)
        candidates_in_column = df[column_name].unique()
        set_candidates.update(candidates_in_column)
    if np.nan in set_candidates:
        set_candidates.remove(np.nan)
    labels_candidates = list(set_candidates)
    d_candidate_index = {candidate: index for index, candidate in enumerate(labels_candidates)}
    n_c = len(labels_candidates)
    # Utilities
    df.replace(d_candidate_index, inplace=True)
    preferences_ut = []
    for index, row in df.iterrows():
        if np.all(np.isnan(row)):
            # Ignore this voter (abstains)
            continue
        voter_utilities = np.zeros(n_c, dtype=int)
        for rank in range(1, last_rank + 1):
            column_name = 'rank{}'.format(rank)
            candidate = row[column_name]
            if not np.isnan(candidate):
                voter_utilities[int(candidate)] = n_c - rank
        preferences_ut.append(voter_utilities)
    np.array(preferences_ut)
    # Convert to Borda format
    preferences_ut = preferences_ut_to_preferences_borda_ut(preferences_ut) - (n_c - 1) / 2
    # Return
    return preferences_ut, labels_candidates



# -*- coding: utf-8 -*-
"""
Created on nov. 03, 2014, 20:36
Copyright François Durand 2014, 2015
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""

import itertools
import numpy as np

from svvamp.preferences.profile import Profile
from svvamp.utils.misc import indent


class StudyProfileCriteria():

    def __init__(self, boolean_criteria=None, numerical_criteria=None,
                 special_candidates_criteria=None, array_criteria=None,
                 matrix_criteria=None):
        """A set of criteria to study for the simulator about the profiles.

        Parameters
        ----------
        boolean_criteria : List of str
            Each string must correspond to a boolean criterion defined in Profile.
        numerical_criteria : List of tuple
            Each tuple is of the form (criterion, func, name), where:
                criterion: the numerical data to be processed.
                func: an aggregating function.
                name: the name chosen for this pair criterion / aggregation function.
        special_candidates_criteria : List of str
            Each string must correspond to an attribute defined in Profile that returns the
            index of a candidate or NaN.
        array_criteria : List of str
            Each string must correspond to a attribute defined in Profile that returns an 1d array of size n_c
            (the number of candidates).
        matrix_criteria : List of str
            Each string must correspond to a attribute defined in Profile that returns an array (whatever
            its size).

        Examples
        --------
        For each argument, if None (default), then a quite extensive pre-defined list is used. Hence, the typical usage
        is simply:

            >>> study_profile_criteria = StudyProfileCriteria()

        Here is an example of manually defined `numerical_criteria`:

            >>> study_profile_criteria = StudyProfileCriteria(numerical_criteria=[
            ...     ('total_utility_mean', np.mean, 'total_u_mean_mean')
            ... ])

        For each profile, this will compute `total_utility_mean` (mean over the candidates). Then it will use `np.mean`
        to calculate the mean over the profiles. The result will be called 'total_u_mean_mean'.
        """
        if boolean_criteria is None:
            boolean_criteria = [
                'exists_condorcet_admissible',
                'exists_weak_condorcet_winner',
                'exists_condorcet_winner_rk_ctb',
                'exists_condorcet_winner_rk',
                'exists_condorcet_winner_ut_rel_ctb',
                'exists_condorcet_winner_ut_rel',
                'exists_condorcet_winner_ut_abs_ctb',
                'exists_condorcet_winner_ut_abs',
                'exists_resistant_condorcet_winner',
                'exists_condorcet_order_rk_ctb',
                'exists_condorcet_order_rk',
                'exists_condorcet_order_ut_rel_ctb',
                'exists_condorcet_order_ut_rel',
                'exists_condorcet_order_ut_abs_ctb',
                'exists_condorcet_order_ut_abs',
                'exists_majority_favorite_rk_ctb',
                'exists_majority_favorite_rk',
                'exists_majority_favorite_ut_ctb',
                'exists_majority_favorite_ut',
                'not_exists_condorcet_admissible',
                'not_exists_weak_condorcet_winner',
                'not_exists_condorcet_winner_rk_ctb',
                'not_exists_condorcet_winner_rk',
                'not_exists_condorcet_winner_ut_rel_ctb',
                'not_exists_condorcet_winner_ut_rel',
                'not_exists_condorcet_winner_ut_abs_ctb',
                'not_exists_condorcet_winner_ut_abs',
                'not_exists_resistant_condorcet_winner',
                'not_exists_condorcet_order_rk_ctb',
                'not_exists_condorcet_order_rk',
                'not_exists_condorcet_order_ut_rel_ctb',
                'not_exists_condorcet_order_ut_rel',
                'not_exists_condorcet_order_ut_abs_ctb',
                'not_exists_condorcet_order_ut_abs',
                'not_exists_majority_favorite_rk_ctb',
                'not_exists_majority_favorite_rk',
                'not_exists_majority_favorite_ut_ctb',
                'not_exists_majority_favorite_ut',
            ]
        self.boolean_criteria = boolean_criteria
        if numerical_criteria is None:
            numerical_criteria = [
                ('mean_utility_min', np.min, 'mean_u_min_min'),
                ('mean_utility_min', np.max, 'mean_u_min_max'),
                ('mean_utility_min', np.mean, 'mean_u_min_mean'),
                ('mean_utility_min', np.std, 'mean_u_min_std'),
                ('mean_utility_max', np.min, 'mean_u_max_min'),
                ('mean_utility_max', np.max, 'mean_u_max_max'),
                ('mean_utility_max', np.mean, 'mean_u_max_mean'),
                ('mean_utility_max', np.std, 'mean_u_max_std'),
                ('mean_utility_mean', np.min, 'mean_u_mean_min'),
                ('mean_utility_mean', np.max, 'mean_u_mean_max'),
                ('mean_utility_mean', np.mean, 'mean_u_mean_mean'),
                ('mean_utility_mean', np.std, 'mean_u_mean_std'),
                ('nb_condorcet_admissible', np.mean,
                 'nb_condorcet_admissible_mean'),
                ('nb_weak_condorcet_winners', np.mean,
                 'nb_weak_condorcet_winners_mean')
            ]
        self.numerical_criteria = numerical_criteria
        if special_candidates_criteria is None:
            special_candidates_criteria = [
                'condorcet_winner_rk_ctb',
                'condorcet_winner_rk',
                'condorcet_winner_ut_rel_ctb',
                'condorcet_winner_ut_rel',
                'condorcet_winner_ut_abs_ctb',
                'condorcet_winner_ut_abs',
                'resistant_condorcet_winner',
                'majority_favorite_rk_ctb',
                'majority_favorite_rk',
                'majority_favorite_ut_ctb',
                'majority_favorite_ut',
            ]
        self.special_candidates_criteria = special_candidates_criteria
        if array_criteria is None:
            array_criteria = [
                'plurality_scores_rk',
                'plurality_scores_ut',
                'condorcet_admissible_candidates',
                'weak_condorcet_winners',
                'borda_score_c_rk',
                'borda_score_c_ut',
                'total_utility_c',
                'mean_utility_c',
                'relative_social_welfare_c',
            ]
        self.array_criteria = array_criteria
        if matrix_criteria is None:
            matrix_criteria = [
                'preferences_ut',
                'preferences_borda_rk',
                'preferences_borda_ut',
                'matrix_duels_ut',
                'matrix_duels_rk',
                'matrix_victories_ut_abs',
                'matrix_victories_ut_abs_ctb',
                'matrix_victories_ut_rel',
                'matrix_victories_ut_rel_ctb',
                'matrix_victories_rk',
                'matrix_victories_rk_ctb'
            ]
        self.matrix_criteria = matrix_criteria

    def __str__(self):
        """
            >>> study_profile_criteria = StudyProfileCriteria(
            ...     boolean_criteria=['exists_condorcet_winner_rk'],
            ...     numerical_criteria=[('mean_utility_min', np.min, 'mean_u_min_min')],
            ...     special_candidates_criteria=['condorcet_winner_rk'],
            ...     array_criteria=['total_utility_c'],
            ...     matrix_criteria=['matrix_duels_rk']
            ... )
            >>> print(study_profile_criteria)  # doctest: +ELLIPSIS
            StudyProfileCriteria with:
                boolean_criteria:
                    exists_condorcet_winner_rk
                numerical_criteria:
                    ('mean_utility_min', <function amin at ...>, 'mean_u_min_min')
                special_candidates_criteria:
                    condorcet_winner_rk
                array_criteria:
                    total_utility_c
                matrix_criteria:
                    matrix_duels_rk

            >>> study_profile_criteria = StudyProfileCriteria(
            ...     boolean_criteria=[],
            ...     numerical_criteria=[],
            ...     special_candidates_criteria=[],
            ...     array_criteria=[],
            ...     matrix_criteria=[]
            ... )
            >>> print(study_profile_criteria)
            StudyProfileCriteria with:
                boolean_criteria: None
                numerical_criteria: None
                special_candidates_criteria: None
                array_criteria: None
                matrix_criteria: None
        """
        # noinspection PyListCreation
        lines = []
        for name in ['boolean_criteria', 'numerical_criteria', 'special_candidates_criteria',
                     'array_criteria', 'matrix_criteria']:
            criteria = getattr(self, name)
            if criteria:
                lines.append(f'{name}:')
                lines.extend(['    ' + str(criterion) for criterion in criteria])
            else:
                lines.append(f'{name}: None')
        return 'StudyProfileCriteria with:\n' + indent('\n'.join(lines))

    def check_sanity(self):
        """Check sanity of the object

        Preforms some basic checks. It is recommended to use this method before launching big simulations.

        Examples
        --------
        Detect an illegal criterion:

            >>> study_profile_criteria = StudyProfileCriteria(boolean_criteria=['unexpected_criterion'])
            >>> study_profile_criteria.check_sanity()
            Traceback (most recent call last):
            ValueError: Attribute 'unexpected_criterion' is unknown for Profile.

        Detect when a aggregation function is not provided:

            >>> study_profile_criteria = StudyProfileCriteria(numerical_criteria=[
            ...     ('mean_utility_min', 'this_is_not_callable', 'mean_u_min_min')
            ... ])
            >>> study_profile_criteria.check_sanity()
            Traceback (most recent call last):
            TypeError: Expected: callable. Got: 'this_is_not_callable'.
        """
        for criterion in itertools.chain(self.boolean_criteria, self.special_candidates_criteria,
                                         self.array_criteria, self.matrix_criteria):
            if criterion not in dir(Profile):
                raise ValueError(f"Attribute {repr(criterion)} is unknown for Profile.")
        for (criterion, func, name) in self.numerical_criteria:
            if criterion not in dir(Profile):
                raise ValueError(f"Attribute {repr(criterion)} is unknown for Profile.")
            if not callable(func):
                raise TypeError(f"Expected: callable. Got: {repr(func)}.")
        print('StudyProfileCriteria: Sanity check was successful.')

    def remove(self, criterion):
        """Remove a criterion

        Parameters
        ----------
        criterion : str

        Examples
        --------
        Remove all occurrences of `criterion` (generally zero or one):

            >>> study_profile_criteria = StudyProfileCriteria(
            ...     boolean_criteria=['exists_condorcet_admissible', 'exists_weak_condorcet_winner'],
            ...     numerical_criteria=[],
            ...     special_candidates_criteria=[],
            ...     array_criteria=[],
            ...     matrix_criteria=[]
            ... )
            >>> study_profile_criteria.remove('exists_weak_condorcet_winner')
            >>> print(study_profile_criteria)
            StudyProfileCriteria with:
                boolean_criteria:
                    exists_condorcet_admissible
                numerical_criteria: None
                special_candidates_criteria: None
                array_criteria: None
                matrix_criteria: None

        For numerical criteria, use the last element of the triple, i.e. the name:

            >>> study_profile_criteria = StudyProfileCriteria(
            ...     boolean_criteria=[],
            ...     numerical_criteria=[('mean_utility_min', np.min, 'mean_u_min_min'),
            ...                         ('mean_utility_min', np.max, 'mean_u_min_max')],
            ...     special_candidates_criteria=[],
            ...     array_criteria=[],
            ...     matrix_criteria=[]
            ... )
            >>> study_profile_criteria.remove('mean_u_min_min')
            >>> print(study_profile_criteria)  # doctest: +ELLIPSIS
            StudyProfileCriteria with:
                boolean_criteria: None
                numerical_criteria:
                    ('mean_utility_min', <function amax at ...>, 'mean_u_min_max')
                special_candidates_criteria: None
                array_criteria: None
                matrix_criteria: None

        """
        for criteria in [self.boolean_criteria, self.special_candidates_criteria,
                         self.array_criteria, self.matrix_criteria]:
            criteria[:] = [c for c in criteria if c != criterion]
        self.numerical_criteria[:] = [c for c in self.numerical_criteria if c[2] != criterion]

    def append_boolean_criterion(self, criterion):
        """Add a boolean criterion

        Parameters
        ----------
        criterion : str

        Examples
        --------
            >>> study_profile_criteria = StudyProfileCriteria(
            ...     boolean_criteria=['exists_condorcet_admissible'],
            ...     numerical_criteria=[],
            ...     special_candidates_criteria=[],
            ...     array_criteria=[],
            ...     matrix_criteria=[]
            ... )
            >>> study_profile_criteria.append_boolean_criterion('exists_weak_condorcet_winner')
            >>> print(study_profile_criteria)
            StudyProfileCriteria with:
                boolean_criteria:
                    exists_condorcet_admissible
                    exists_weak_condorcet_winner
                numerical_criteria: None
                special_candidates_criteria: None
                array_criteria: None
                matrix_criteria: None
        """
        if criterion not in self.boolean_criteria:
            self.boolean_criteria.append(criterion)

    def append_numerical_criterion(self, criterion, func, name):
        """Add a numerical criterion

        Parameters
        ----------
        criterion : str
            The numerical data to be processed.
        func : callable
            An aggregating function.
        name : str
            The name chosen for this pair criterion / aggregation function.

        Examples
        --------
            >>> study_profile_criteria = StudyProfileCriteria(
            ...     boolean_criteria=[],
            ...     numerical_criteria=[],
            ...     special_candidates_criteria=[],
            ...     array_criteria=[],
            ...     matrix_criteria=[]
            ... )
            >>> study_profile_criteria.append_numerical_criterion(
            ...     'total_utility_mean', np.mean, 'total_u_mean_mean'
            ... )
            >>> print(study_profile_criteria)  # doctest: +ELLIPSIS
            StudyProfileCriteria with:
                boolean_criteria: None
                numerical_criteria:
                    ('total_utility_mean', <function mean at ...>, 'total_u_mean_mean')
                special_candidates_criteria: None
                array_criteria: None
                matrix_criteria: None

        For each profile, this will compute `total_utility_mean` (mean over the candidates). Then use `np.mean` to
        calculate the mean over the profiles. The result will be called 'total_u_mean_mean'.
        """
        if name not in [n for (c, f, n) in self.numerical_criteria]:
            self.numerical_criteria.append((criterion, func, name))

    def rename_numerical_criterion(self, old_name, new_name):
        """Rename a numerical criterion

        Parameters
        ----------
        old_name : str
        new_name : str

        Examples
        --------
            >>> study_profile_criteria = StudyProfileCriteria(
            ...     boolean_criteria=[],
            ...     numerical_criteria=[('total_utility_mean', np.mean, 'total_u_mean_mean')],
            ...     special_candidates_criteria=[],
            ...     array_criteria=[],
            ...     matrix_criteria=[]
            ... )
            >>> study_profile_criteria.rename_numerical_criterion('total_u_mean_mean', 'the_new_name')
            >>> print(study_profile_criteria)  # doctest: +ELLIPSIS
            StudyProfileCriteria with:
                boolean_criteria: None
                numerical_criteria:
                    ('total_utility_mean', <function mean at ...>, 'the_new_name')
                special_candidates_criteria: None
                array_criteria: None
                matrix_criteria: None
        """
        self.numerical_criteria[:] = [
            (c, f, new_name) if n == old_name else (c, f, n)
            for (c, f, n) in self.numerical_criteria
        ]

    def append_special_candidates_criterion(self, criterion):
        """Add a 'special candidates' criterion

        Parameters
        ----------
        criterion : str

        Examples
        --------
            >>> study_profile_criteria = StudyProfileCriteria(
            ...     boolean_criteria=[],
            ...     numerical_criteria=[],
            ...     special_candidates_criteria=['condorcet_winner_rk_ctb'],
            ...     array_criteria=[],
            ...     matrix_criteria=[]
            ... )
            >>> study_profile_criteria.append_special_candidates_criterion('condorcet_winner_rk')
            >>> print(study_profile_criteria)
            StudyProfileCriteria with:
                boolean_criteria: None
                numerical_criteria: None
                special_candidates_criteria:
                    condorcet_winner_rk_ctb
                    condorcet_winner_rk
                array_criteria: None
                matrix_criteria: None
        """
        if criterion not in self.special_candidates_criteria:
            self.special_candidates_criteria.append(criterion)

    def append_array_criterion(self, criterion):
        """Add an array criterion

        Parameters
        ----------
        criterion : str

        Examples
        --------
            >>> study_profile_criteria = StudyProfileCriteria(
            ...     boolean_criteria=[],
            ...     numerical_criteria=[],
            ...     special_candidates_criteria=[],
            ...     array_criteria=['plurality_scores_rk'],
            ...     matrix_criteria=[]
            ... )
            >>> study_profile_criteria.append_array_criterion('plurality_scores_ut')
            >>> print(study_profile_criteria)
            StudyProfileCriteria with:
                boolean_criteria: None
                numerical_criteria: None
                special_candidates_criteria: None
                array_criteria:
                    plurality_scores_rk
                    plurality_scores_ut
                matrix_criteria: None
        """
        if criterion not in self.array_criteria:
            self.array_criteria.append(criterion)

    def append_matrix_criterion(self, criterion):
        """Add a matrix criterion

        Parameters
        ----------
        criterion : str

        Examples
        --------
            >>> study_profile_criteria = StudyProfileCriteria(
            ...     boolean_criteria=[],
            ...     numerical_criteria=[],
            ...     special_candidates_criteria=[],
            ...     array_criteria=[],
            ...     matrix_criteria=['preferences_ut']
            ... )
            >>> study_profile_criteria.append_matrix_criterion('preferences_borda_rk')
            >>> print(study_profile_criteria)
            StudyProfileCriteria with:
                boolean_criteria: None
                numerical_criteria: None
                special_candidates_criteria: None
                array_criteria: None
                matrix_criteria:
                    preferences_ut
                    preferences_borda_rk
        """
        if criterion not in self.matrix_criteria:
            self.matrix_criteria.append(criterion)



from svvamp import RuleExhaustiveBallot


def test_fast_algo_setter():
    """
        >>> rule = RuleExhaustiveBallot()
        >>> rule.fast_algo = 'foo'
        Traceback (most recent call last):
        ValueError: Unknown value for option fast_algo: foo
    """
    pass


def test_log_um_():
    """
        >>> rule = RuleExhaustiveBallot(um_option='exact')
        >>> rule.log_um_
        'um_option = exact'
    """
    pass


def test_log_cm_():
    """
        >>> rule = RuleExhaustiveBallot(cm_option='exact')
        >>> rule.log_cm_
        'cm_option = exact'
    """
    pass



# -*- coding: utf-8 -*-
"""
Created on 10 dec. 2018, 16:53
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from math import floor, ceil
from svvamp.rules.rule import Rule
from svvamp.utils import type_checker
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile


class RuleRangeVoting(Rule):
    """Range Voting with Average.

    Parameters
    ----------
    min_grade : number
        Minimal grade allowed.
    max_grade : number
        Maximal grade allowed.
    step_grade : number
        Interval between two consecutive allowed grades.

        * If ``step_grade = 0``, all grades in the interval [:attr:`min_grade`, :attr:`max_grade`] are allowed
          ('continuous' set of grades).
        * If ``step_grade > 0``, authorized grades are the multiples of :attr:`step_grade` lying in the interval
          [:attr:`min_grade`, :attr:`max_grade`]. In addition, the grades :attr:`min_grade` and :attr:`max_grade` are
          always authorized, even if they are not multiples of :attr:`step_grade`.

    rescale_grades : bool
        Whether sincere voters rescale their utilities to produce grades.

        * If ``rescale_grades`` = ``True``, then each sincere voter ``v`` applies an affine transformation to send
          her utilities into the interval [:attr:`min_grade`, :attr:`max_grade`].
        * If ``rescale_grades`` = ``False``, then each sincere voter ``v`` clips her utilities into the interval
          [:attr:`min_grade`, :attr:`max_grade`].

        See :attr:`ballots` for more details.

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleRangeVoting()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[1.   0.5  0.  ]
         [1.   0.   0.75]
         [1.   1.   0.  ]
         [0.5  0.   1.  ]
         [0.   0.   1.  ]]
        scores =
        [0.7  0.3  0.55]
        candidates_by_scores_best_to_worst
        [0 2 1]
        scores_best_to_worst
        [0.7  0.55 0.3 ]
        w = 0
        score_w = 0.7
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  False
        is_based_on_rk =  False
        is_based_on_ut_minus1_1 =  True
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (False) ==> Condorcet_c_rk (False)     ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (False)     ==>     Condorcet_ut_abs_c (False)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (False) ==> maj_fav_c_rk (False)       ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (False) ==> majority_favorite_c_ut (False)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (True)                 ==>                IgnMC_c (True)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (True)                 ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = False
        log_im: im_option = exact
        candidates_im =
        [0. 0. 0.]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = False
        log_um: um_option = exact
        candidates_um =
        [0. 0. 0.]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = exact
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 6. 4.]
        sufficient_coalition_size_icm =
        [0. 6. 4.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = False
        log_cm: cm_option = exact
        candidates_cm =
        [0. 0. 0.]
        necessary_coalition_size_cm =
        [0. 3. 3.]
        sufficient_coalition_size_cm =
        [0. 3. 3.]

    Notes
    -----
    Each voter attributes a grade to each candidate. By default, authorized grades are all numbers in the interval
    [:attr:`min_grade`, :attr:`max_grade`]. To use a discrete set of notes, modify attribute :attr:`step_grade`.

    The candidate with highest average grade wins. In case of a tie, the tied candidate with lowest index is declared
    the winner.

    Default behavior of sincere voters: voter ``v`` applies an affine transformation to her utilities
    :attr:`preferences_ut`\ ``[v, :]`` to get her grades, such that her least-liked candidate receives
    :attr:`min_grade` and her most-liked candidate receives :attr:`max_grade`. To modify this behavior, use attribute
    :attr:`rescale_grades`. For more details about the behavior of sincere voters, see :attr:`ballots`.

    * :meth:`~svvamp.Election.not_iia`:

        * If :attr:`rescale_grades` = ``False``, then Range voting always meets IIA.
        * If :attr:`rescale_grades` = ``True``, then non-polynomial or non-exact algorithms from superclass
          :class:`Rule` are used.

    * :meth:`is_cm_`, :meth:`is_icm_`, :meth:`is_im_`, :meth:`is_tm_`, :meth:`is_um_`: Exact in polynomial time.
    """

    full_name = 'Range Voting'
    abbreviation = 'RV'

    options_parameters = Rule.options_parameters.copy()
    options_parameters.update({
        'max_grade': {'allowed': np.isfinite, 'default': 1},
        'min_grade': {'allowed': np.isfinite, 'default': 0},
        'step_grade': {'allowed': np.isfinite, 'default': 0},
        'rescale_grades': {'allowed': type_checker.is_bool, 'default': True},
        'im_option': {'allowed': ['exact'], 'default': 'exact'},
        'tm_option': {'allowed': ['exact'], 'default': 'exact'},
        'um_option': {'allowed': ['exact'], 'default': 'exact'},
        'icm_option': {'allowed': ['exact'], 'default': 'exact'},
        'cm_option': {'allowed': ['exact'], 'default': 'exact'}
    })

    def __init__(self, **kwargs):
        self._min_grade = None
        self._max_grade = None
        self._step_grade = None
        self._rescale_grades = None
        super().__init__(
            with_two_candidates_reduces_to_plurality=False,
            # # Even if ``rescale_grades = True``, a voter who has the same utility for ``c`` and ``d`` will not vote
            # # the same in Range Voting and in Plurality.
            is_based_on_rk=False,
            precheck_um=False, precheck_tm=False, precheck_icm=False,
            log_identity="RANGE_VOTING", **kwargs
        )

    # %% Setting the parameters

    @property
    def min_grade(self):
        return self._min_grade

    @min_grade.setter
    def min_grade(self, value):
        if self._min_grade == value:
            return
        if self.options_parameters['min_grade']['allowed'](value):
            self.mylogv("Setting min_grade =", value, 1)
            self._min_grade = value
            self._result_options['min_grade'] = value
            self.delete_cache()
        else:
            raise ValueError("Unknown value for min_grade: " + format(value))

    @property
    def max_grade(self):
        return self._max_grade

    @max_grade.setter
    def max_grade(self, value):
        if self._max_grade == value:
            return
        if self.options_parameters['max_grade']['allowed'](value):
            self.mylogv("Setting max_grade =", value, 1)
            self._max_grade = value
            self._result_options['max_grade'] = value
            self.delete_cache()
        else:
            raise ValueError("Unknown value for max_grade: " + format(value))

    @property
    def step_grade(self):
        return self._step_grade

    @step_grade.setter
    def step_grade(self, value):
        if self._step_grade == value:
            return
        if self.options_parameters['step_grade']['allowed'](value):
            self.mylogv("Setting step_grade =", value, 1)
            self._step_grade = value
            self._result_options['step_grade'] = value
            self.delete_cache()
        else:
            raise ValueError("Unknown value for step_grade: " + format(value))

    @property
    def rescale_grades(self):
        return self._rescale_grades

    @rescale_grades.setter
    def rescale_grades(self, value):
        if self._rescale_grades == value:
            return
        if self.options_parameters['rescale_grades']['allowed'](value):
            self.mylogv("Setting rescale_grades =", value, 1)
            self._rescale_grades = value
            self._result_options['rescale_grades'] = value
            self.delete_cache()
        else:
            raise ValueError("Unknown value for rescale_grades: " + format(value))

    @cached_property
    def allowed_grades(self):
        """List or None. If ``step_grade`` is positive, the list of authorized grades. If ``step_grade`` is zero
        (continuous set of grades), then None.
        """
        if self.step_grade == 0:
            return None
        else:
            i_lowest_rung = floor(self.min_grade / self.step_grade) + 1
            i_highest_rung = ceil(self.max_grade / self.step_grade) - 1
            return np.concatenate((
                [self.min_grade],
                np.array(range(i_lowest_rung, i_highest_rung + 1)) * self.step_grade,
                [self.max_grade]
            ))

    # %% Counting the ballots

    @cached_property
    def ballots_(self):
        """2d array of integers. ``ballots[v, c]`` is the grade attributed by voter ``v`` to candidate ``c`` (when
        voting sincerely). The following process is used.

        1. Convert utilities into grades in the interval [:attr:`min_grade`, :attr:`max_grade`].

            * If :attr:`rescale_grades` = ``True``, then each voter ``v`` applies an affine transformation to
              :attr:`preferences_ut`\ ``[v, :]`` such that her least-liked candidate receives :attr:`min_grade` and
              her most-liked candidate receives :attr:`max_grade`. Exception: if she is indifferent between all
              candidates, then she attributes (:attr:`min_grade` + :attr:`max_grade`) / 2 to all of them.

            * If :attr:`rescale_grades` = ``False``, then each voter ``v`` clips her utilities into the interval
              [:attr:`min_grade`, :attr:`max_grade`]: each utility greater than :attr:`max_grade` (resp. lower than
              :attr:`min_grade`) becomes :attr:`max_grade` (resp. :attr:`min_grade`).

        2. If :attr:`step_grades` > 0 (discrete set of grades), round each grade to the closest authorized grade.
        """
        self.mylog("Compute ballots", 1)
        # Rescale (or not)
        if self.rescale_grades:
            # Multiplicative renormalization
            max_util = np.max(self.profile_.preferences_ut, axis=1)
            min_util = np.min(self.profile_.preferences_ut, axis=1)
            delta_util = max_util - min_util
            ballots = np.divide(
                self.profile_.preferences_ut * (self.max_grade - self.min_grade),
                delta_util[:, np.newaxis],
                out=np.zeros(self.profile_.preferences_ut.shape),
                where=delta_util[:, np.newaxis] != 0
            )  # `out` and `where` ensure that when dividing by zero, the result will be zero.
            # Additive renormalization
            middle_grade = (np.max(ballots, axis=1) + np.min(ballots, axis=1)) / 2
            ballots += (self.max_grade - self.min_grade) / 2 - middle_grade[:, np.newaxis]
        else:
            ballots = np.clip(self.profile_.preferences_ut, self.min_grade, self.max_grade)
        # Round (or not)
        if self.step_grade != 0:
            if self.min_grade % 1 == 0 and self.max_grade % 1 == 0 and self.step_grade % 1 == 0:
                # There is a faster version for this (common) use case.
                ballots = np.array(np.rint(ballots), dtype=int)
            else:
                frontiers = (self.allowed_grades[0:-1] + self.allowed_grades[1:]) / 2
                for v in range(self.profile_.n_v):
                    ballots[v, :] = self.allowed_grades[np.digitize(ballots[v, :], frontiers)]
        return ballots

    @cached_property
    def scores_sum_(self):
        """1d array of floats. ``scores_sum_[c]`` is the total grade of candidate ``c``.
        """
        # Note for developers: this is used internally for manipulation computations. Compared to the average grade,
        # this limits problems of float approximation, especially for IM method.
        self.mylog("Compute scores_sum_", 1)
        return np.sum(self.ballots_, 0)

    @cached_property
    def scores_(self):
        """1d array of floats. ``scores_[c]`` is the average grade of candidate ``c``.
        """
        return np.around(self.scores_sum_ / self.profile_.n_v, 12)

    # %% Manipulation criteria of the voting system

    @cached_property
    def meets_ignmc_c_ctb(self):
        return True

    # %% Independence of Irrelevant Alternatives (IIA)

    @property
    def meets_iia(self):
        return not self.rescale_grades

    @property
    def is_based_on_ut_minus1_1(self):
        return self._rescale_grades

    @meets_iia.setter
    def meets_iia(self, value):
        pass

    @is_based_on_ut_minus1_1.setter
    def is_based_on_ut_minus1_1(self, value):
        pass

    # %% Individual manipulation (IM)

    def _im_main_work_v_(self, v, c_is_wanted, nb_wanted_undecided, stop_if_true):
        """
            >>> profile = Profile([
            ...     [ 1. , -0.5,  0. ],
            ...     [ 0.5,  1. , -1. ],
            ...     [-0.5,  0.5, -1. ],
            ...     [ 1. ,  0. ,  1. ],
            ...     [-1. , -0.5,  1. ],
            ... ])
            >>> rule = RuleRangeVoting()(profile)
            >>> rule.is_im_c_(1)
            False

            >>> profile = Profile([
            ...     [ 1. , -1. ,  1. ],
            ...     [ 1. , -1. ,  0.5],
            ...     [-0.5,  1. ,  0.5],
            ...     [-1. ,  0. , -0.5],
            ...     [-1. ,  0.5, -0.5],
            ... ])
            >>> rule = RuleRangeVoting()(profile)
            >>> rule.is_im_c_(1)
            True
        """
        scores_without_v = self.scores_sum_ - self.ballots_[v, :]
        w_without_v = np.argmax(np.around(scores_without_v, 12))
        new_score_w = np.around(scores_without_v[w_without_v] + self.min_grade, 12)
        for c in range(self.profile_.n_c):
            if not c_is_wanted[c]:
                continue
            if not np.isneginf(self._v_im_for_c[v, c]):
                continue
            new_score_c = np.around(scores_without_v[c] + self.max_grade, 12)
            if (c < w_without_v and new_score_c >= new_score_w) or (c > w_without_v and new_score_c > new_score_w):
                self._v_im_for_c[v, c] = True
                self._candidates_im[c] = True
                self._voters_im[v] = True
                self._is_im = True
                self.mylog("IM found", 3)
                if stop_if_true:
                    return
            else:
                self._v_im_for_c[v, c] = False
            nb_wanted_undecided -= 1
            if nb_wanted_undecided == 0:
                return

    # %% Coalition Manipulation (CM)

    def _cm_main_work_c_(self, c, optimize_bounds):
        scores_temp = np.sum(self.ballots_[np.logical_not(self.v_wants_to_help_c_[:, c]), :], 0)
        w_temp = np.argmax(np.around(scores_temp, 12))
        sufficient_not_rounded = np.around((scores_temp[w_temp] - scores_temp[c]) / (self.max_grade - self.min_grade),
                                           12)
        self.mylogv("sufficient_not_rounded =", sufficient_not_rounded, 2)
        if sufficient_not_rounded % 1 == 0 and c > w_temp:
            self._sufficient_coalition_size_cm[c] = sufficient_not_rounded + 1
        else:
            self._sufficient_coalition_size_cm[c] = ceil(sufficient_not_rounded)
        self._necessary_coalition_size_cm[c] = self._sufficient_coalition_size_cm[c]

    # %% Trivial Manipulation (TM)

    @cached_property
    def is_tm_(self):
        return self.is_cm_

    def is_tm_c_(self, c):
        return self.is_cm_c_(c)

    @cached_property
    def candidates_tm_(self):
        return self.candidates_cm_

    # %% Unison Manipulation (UM)

    @cached_property
    def is_um_(self):
        return self.is_cm_

    def is_um_c_(self, c):
        return self.is_cm_c_(c)

    @cached_property
    def candidates_um_(self):
        return self.candidates_cm_

    # %% Ignorant-Coalition Manipulation (ICM)

    # Since Range Voting meets IgnMC_c_ctb, the general methods are exact.



# -*- coding: utf-8 -*-
"""
Created on 11 dec. 2018, 07:50
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
import networkx as nx
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile


class RuleRankedPairs(Rule):
    """Tideman's Ranked Pairs.

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleRankedPairs()(profile)
        >>> rule.meets_condorcet_c_rk_ctb
        True
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        scores =
        [[0. 3. 3.]
         [0. 0. 0.]
         [0. 3. 0.]]
        candidates_by_scores_best_to_worst
        [0, 2, 1]
        scores_best_to_worst
        [[0. 3. 3.]
         [0. 0. 3.]
         [0. 0. 0.]]
        w = 0
        score_w = [[0. 3. 3.]]
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (True)  ==> Condorcet_c_rk (True)      ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (True)      ==>     Condorcet_ut_abs_c (True)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (True)  ==> maj_fav_c_rk (True)        ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (True)  ==> majority_favorite_c_ut (True)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (True)                 ==>                IgnMC_c (True)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (True)                 ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = nan
        log_im: im_option = lazy
        candidates_im =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = nan
        log_um: um_option = lazy
        candidates_um =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = exact
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 6. 4.]
        sufficient_coalition_size_icm =
        [0. 6. 4.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = nan
        log_cm: cm_option = lazy, um_option = lazy, tm_option = exact
        candidates_cm =
        [ 0.  0. nan]
        necessary_coalition_size_cm =
        [0. 1. 2.]
        sufficient_coalition_size_cm =
        [0. 6. 4.]

    Notes
    -----
    In the matrix of duels :attr:`matrix_duels_rk`, victories (and ties) are sorted by decreasing amplitude. If two
    duels have the same score, we take first the one where the winner has the smallest index; if there is still a
    choice to make, we take first the duel where the loser has the highest index.

    Starting with the largest victory, we build a directed graph whose nodes are the candidates and edges are
    victories. But if a victory creates a cycle in the graph, it is not validated and the edge is not added.

    At the end, we have a transitive connected directed graph, whose adjacency relation is included in the relation
    of victories (with ties broken), :attr:`matrix_victories_rk_ctb`. The maximal node of this graph (by topological
    order) is declared the winner.

    This method meets the Condorcet criterion.

    * :meth:`is_cm_`: Deciding CM is NP-complete. Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_icm_`: Exact in polynomial time.
    * :meth:`is_im_`: Deciding IM is NP-complete. Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_iia_`: Exact in polynomial time.
    * :meth:`is_tm_`: Exact in polynomial time.
    * :meth:`is_um_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.

    References
    ----------
    'Independence of clones as a criterion for voting rules', Nicolaus Tideman, 1987.

    'Complexity of Unweighted Coalitional Manipulation under Some Common Voting Rules', Lirong Xia et al., 2009.

    'Schulze and Ranked-Pairs Voting are Fixed-Parameter Tractable to Bribe, Manipulate, and Control',
    Lane A. Hemaspaandra, Rahman Lavaee and Curtis Menton, 2012.

    'A Complexity-of-Strategic-Behavior Comparison between Schulze’s Rule and Ranked Pairs', David Parkes and
    Lirong Xia, 2012.
    """

    full_name = 'Ranked Pairs'
    abbreviation = 'RP'

    options_parameters = Rule.options_parameters.copy()
    options_parameters['icm_option'] = {'allowed': ['exact'], 'default': 'exact'}

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_icm=False,
            log_identity="RANKED_PAIRS", **kwargs
        )

    # %% Counting the ballots

    @cached_property
    def _count_ballots_(self):
        self.mylog("Count ballots", 1)
        matrix_duels_vtb_temp = np.copy(self.profile_.matrix_duels_rk)
        dig = nx.DiGraph()
        dig.add_nodes_from(range(self.profile_.n_c))
        while True:
            best_duel_result_c = np.max(matrix_duels_vtb_temp, 1)
            c = np.argmax(best_duel_result_c)
            best_duel_result = best_duel_result_c[c]
            d = np.where(matrix_duels_vtb_temp[c, :] == best_duel_result)[0][-1]
            if best_duel_result == 0:
                break
            matrix_duels_vtb_temp[c, d] = 0
            matrix_duels_vtb_temp[d, c] = 0
            if not nx.has_path(dig, d, c):
                dig.add_edge(c, d, weight=self.profile_.matrix_duels_rk[c, d])
        candidates_by_scores_best_to_worst = list(nx.topological_sort(dig))
        w = candidates_by_scores_best_to_worst[0]
        scores = nx.to_numpy_matrix(dig)
        return {'scores': scores, 'w': w, 'candidates_by_scores_best_to_worst': candidates_by_scores_best_to_worst}

    @cached_property
    def w_(self):
        return self._count_ballots_['w']

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        """1d array of integers. ``candidates_by_scores_best_to_worst[k]`` is the ``k``\ :sup:`th` candidate by
        topological order on the graph generated by Ranked Pairs.
        """
        return self._count_ballots_['candidates_by_scores_best_to_worst']

    @cached_property
    def scores_(self):
        """2d array of integers. ``scores[c, d]`` is equal to :attr:`matrix_duels_rk`\ ``[c, d]`` iff this duel was
        validated in Ranked Pairs, 0 otherwise.

        .. note::

            Unlike for most other voting systems, ``scores`` matrix must be read in rows, in order to comply with our
            convention for the matrix of duels: ``c``'s score vector is ``scores[c, :]``.
        """
        return self._count_ballots_['scores']

    @cached_property
    def score_w_(self):
        """1d array. ``score_w_`` is :attr:`w_`'s score vector: ``score_w_`` =
        :attr:`scores_`\ ``[``:attr:`w_`\ ``, :]``.
        """
        self.mylog("Compute winner's score", 1)
        return self.scores_[self.w_, :]

    @cached_property
    def scores_best_to_worst_(self):
        """2d array. ``scores_best_to_worst_`` is the scores of the candidates, from the winner to the last candidate
        of the election.

        ``scores_best_to_worst[k, j]`` is the score of the ``k``\ :sup:`th` best candidate of the election against
        the ``j``\ :sup:`th`. It is the result in :attr:`~svvamp.Population.matrix_duels_rk` if this duels was
        validated by Ranked Pairs, 0 otherwise.
        """
        self.mylog("Compute scores_best_to_worst", 1)
        return self.scores_[self.candidates_by_scores_best_to_worst_, :][:, self.candidates_by_scores_best_to_worst_]

    @cached_property
    def v_might_im_for_c_(self):
        # This is a quite lazy version, we could be more precise (possible future work). For a voter to be pivotal,
        # it is necessary that she makes a duel examined before another one in Ranked Pairs algorithm (this includes
        # the case of making a duel ``(c, d)`` examined before ``(d, c)``, i.e. changing a defeat into a victory)
        self.mylog("Compute v_might_im_for_c_", 1)
        scores_duels = np.sort(np.concatenate((
            self.profile_.matrix_duels_rk[np.triu_indices(self.profile_.n_c, 1)],
            self.profile_.matrix_duels_rk[np.tril_indices(self.profile_.n_c, -1)]
        )))
        one_v_might_be_pivotal = np.any(scores_duels[:-1] + 2 >= scores_duels[1:])
        return np.full((self.profile_.n_v, self.profile_.n_c), one_v_might_be_pivotal)

    # %% Manipulation criteria of the voting system

    @cached_property
    def meets_condorcet_c_rk_ctb(self):
        return True
